/***
 * OW2 FraSCAti Assembly Factory
 * Copyright (C) 2008-2009 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org 
 *
 * Author: Damien Fournier
 */

package org.ow2.frascati.factory.core.domain;

import java.util.Collection;
import java.util.HashMap;
import java.util.Set;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.ContentController;
import org.objectweb.fractal.api.control.IllegalContentException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.fractal.api.control.NameController;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.util.Fractal;
import org.ow2.frascati.factory.core.domain.api.CompositeDomain;
import org.ow2.frascati.tinfi.TinfiDomain;

/**
 * Manage SCA components into Fractal composite (Domain)
 * 
 * @author <a href="mailto:damien.fournier@inria.fr">Damien Fournier</a>
 * @version $Revision$
 * @since 0.6
 */

@org.objectweb.fractal.fraclet.annotations.Component
public class CompositeDomainImpl implements CompositeDomain {

  @Requires
  protected Component domain;

  private HashMap<String, Component> index = new HashMap<String, Component>();

  /**
   * Register component to composite domain
   * 
   * @param component
   * @throws DomainException 
   */

  public void add(Component component) throws DomainException {

    try {
      // Retrieve Component name
      NameController nc = org.objectweb.fractal.util.Fractal
          .getNameController(component);
      String componentName = nc.getFcName();

      if (!index.containsKey(componentName)) {
        Fractal.getContentController(this.domain).addFcSubComponent(component);
        index.put(componentName, component);
      } else {
        throw new DomainException("Component " + componentName + " already loaded into domain");
      }

    } catch (NoSuchInterfaceException e) {
      throw new DomainException("Could not retrieve domain controller interface",e);
    } catch (IllegalContentException e) {
      throw new DomainException("Could not retrieve domain content",e);
    } catch (IllegalLifeCycleException e) {
      throw new DomainException(e);
    }
  }

  public Component get(String name) {
    return index.get(name);
  }

  /**
   * Remove component from Composite Domain
   * 
   * @param name
   *          - Name of the component to remove
   * @throws DomainException 
   */

  public void remove(String name) throws DomainException {

    // Retrieve component reference from its name
    Component component = index.get(name);
    
    if (component == null) throw new DomainException("Component " + name +" does not exist");

    try {
      
      LifeCycleController domainLifeCycle = Fractal.getLifeCycleController(domain);
      ContentController domainContent = Fractal.getContentController(domain);
      
      // Stop Domain
      domainLifeCycle.stopFc();
      // Remove component
      domainContent.removeFcSubComponent(component);
      // Start Domain
      domainLifeCycle.startFc();
      
      // Remove component from index
      index.remove(name);
      
      // Use tinfi to close SCA component
      TinfiDomain.close(component);
      
    } catch (IllegalContentException e) {
      throw new DomainException("Could not retrieve domain content",e);
    } catch (IllegalLifeCycleException e) {
      throw new DomainException(e);
    } catch (NoSuchInterfaceException e) {
      throw new DomainException("Could not retrieve domain controller interface",e);
    }

  }

  /**
   * List component names for this composite domain
   * 
   * @return
   */

  public String[] list() {
    Set<String> keyset = index.keySet();
    return keyset.toArray(new String[keyset.size()]);
  }

  /**
   * Update index for composite domain
   * @throws DomainException 
   */

  public void index() throws DomainException {
    HashMap<String, Component> newIndex = new HashMap<String, Component>();
    try {
      Component[] components = Fractal.getContentController(this.domain)
          .getFcSubComponents();
      for (Component component : components) {
        String componentName = Fractal.getNameController(component).getFcName();
        newIndex.put(componentName, component);
      }
      this.index = newIndex;
    } catch (NoSuchInterfaceException e) {
      throw new DomainException("Could not build index for composite domain",e);
    }
  }

  /* (non-Javadoc)
   * @see org.ow2.frascati.factory.core.domain.api.CompositeDomain#getAll()
   */
  
  public Component[] getAll() {
    Collection<Component> components = index.values();
    return components.toArray(new Component [components.size()] );
  }

}
