/***
 * OW2 FraSCAti Assembly Factory
 * Copyright (C) 2008-2009 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org 
 *
 * Author: Damien Fournier
 * 
 * Contributor(s): Nicolas Dolet
 * 
 */

package org.ow2.frascati.factory.core.instance.component;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.stp.sca.BaseReference;
import org.eclipse.stp.sca.BaseService;
import org.eclipse.stp.sca.Component;
import org.eclipse.stp.sca.ComponentReference;
import org.eclipse.stp.sca.ComponentService;
import org.eclipse.stp.sca.Composite;
import org.eclipse.stp.sca.Property;
import org.eclipse.stp.sca.PropertyValue;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalContentException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.control.NameController;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.fraclet.types.Contingency;
import org.objectweb.fractal.util.Fractal;
import org.ow2.frascati.factory.core.instance.NotSupportedException;
import org.ow2.frascati.factory.core.instance.binding.BindingException;
import org.ow2.frascati.factory.core.instance.binding.ScaBindingProcess;
import org.ow2.frascati.factory.core.instance.implementation.ImplementationException;
import org.ow2.frascati.factory.core.instance.implementation.ScaImplementation;
import org.ow2.frascati.factory.core.instance.intent.IntentException;
import org.ow2.frascati.factory.core.instance.intent.ScaIntent;
import org.ow2.frascati.factory.core.instance.interfaces.InterfaceException;
import org.ow2.frascati.factory.core.instance.interfaces.ScaInterface;
import org.ow2.frascati.factory.core.instance.property.ScaProperty;
import org.ow2.frascati.factory.core.instance.runtime.RuntimeFactory;

/**
 * Create components instances
 * 
 * @author <a href="mailto:damien.fournier@inria.fr">Damien Fournier</a>
 * @version $Revision$
 */

@org.objectweb.fractal.fraclet.annotations.Component
public class ScaComponentImpl implements ScaComponent {

  /**
   * Logger for this class
   */

  java.util.logging.Logger log = java.util.logging.Logger
      .getLogger(ScaComponentImpl.class.getCanonicalName());

  /**
   * The required runtime
   */

  @Requires(name = "runtime")
  RuntimeFactory runtime;

  /**
   * The implementation switcher component
   */

  @Requires(name = "implementation")
  ScaImplementation implementation;

  /**
   * The java interface builder component
   */

  @Requires(name = "component-interface")
  ScaInterface itf;

  /**
   * The property setter component
   */

  @Requires(name = "property", contingency = Contingency.OPTIONAL)
  ScaProperty property;

  /**
   * The required scaBindingProcess constructor component
   */

  @Requires(name = "binding", contingency = Contingency.OPTIONAL)
  ScaBindingProcess scaBindingProcess;

  /**
   * The intent resolver component
   */

  @Requires(name = "intent", contingency = Contingency.OPTIONAL)
  ScaIntent intents;

  public org.objectweb.fractal.api.Component create(Component comp)
      throws ImplementationException, NotSupportedException, InterfaceException {
    log.fine("Building component " + comp.getName());

    org.objectweb.fractal.api.Component component;
    ComponentType componentType;

    try {
      componentType = create_componentType(comp);
      log.finer("Create component " + comp.getName() + " instance");
      component = implementation
          .create(comp.getImplementation(), componentType);
      
      // can only be done if component for sca intent and component instance exist
      if (component != null) {
        // set the name for this component
        NameController nc = Fractal.getNameController(component);
        nc.setFcName(comp.getName());
        // Initialize intents for the component, its references and its services
        initIntents(component, comp);
      }
      return component;
    } catch (NoSuchInterfaceException e) {
      e.printStackTrace();
    } catch (IntentException e) {
      e.printStackTrace();
    }
    return null;
  }

  public org.objectweb.fractal.api.Component create(Composite assembly)
      throws ImplementationException, NotSupportedException, BindingException,
      InterfaceException {

    org.objectweb.fractal.api.Component assemblyComponent = null;
    ComponentType assemblyType = null;

    try {

      // Create container
      ComponentType containerType = runtime.createFcType(new InterfaceType[0]);
      org.objectweb.fractal.api.Component container = runtime.newFcInstance(
          containerType, new Object[] { runtime.getClassLoader(),
              "scaComposite" }, null);

      // Create assembly
      assemblyType = create_componentType(assembly);
      assemblyComponent = runtime.newFcInstance(assemblyType, new Object[] {
          runtime.getClassLoader(), "scaComposite" }, null);

      // Registering components while their are created, improve performance
      Map<String, org.objectweb.fractal.api.Component> componentMap = new HashMap<String, org.objectweb.fractal.api.Component>();

      // Create assembly components and add them to the assembly
      EList<org.eclipse.stp.sca.Component> components = assembly.getComponent();
      for (org.eclipse.stp.sca.Component c : components) {

        org.objectweb.fractal.api.Component newComponent = create(c);

        // Add Component to the assembly only if a component instance is
        // returned
        if (newComponent != null) {
          componentMap.put(Fractal.getNameController(newComponent).getFcName(),
              newComponent);
          Fractal.getContentController(assemblyComponent).addFcSubComponent(
              newComponent);

          // set properties if applicable, can only be done once components have
          // been added to the current composite

          for (PropertyValue aProperty : c.getProperty())
            property.setProperty(aProperty, newComponent);

        }

      }

      // Can only be done if container exist
      if (container!=null) {

        // set the name for the assembly container
        NameController nc = Fractal.getNameController(container);
        nc.setFcName(assembly.getName() + "-container");

        // set the name for this assembly
        NameController na = Fractal.getNameController(assemblyComponent);
        na.setFcName(assembly.getName());

        // Add the SCA assembly into the container
        Fractal.getContentController(container).addFcSubComponent(
            assemblyComponent);

        // set properties of the assembly
        for (Property aProperty : assembly.getProperty())
          property.setProperty(aProperty, assemblyComponent);

        // TODO Required with binding factory 0.5
        // Save current context ClassLoader
        // ClassLoader contextClassLoader =
        // Thread.currentThread().getContextClassLoader();
        // // Test if context ClassLoader is different from classloader used by
        // AF
        // if (contextClassLoader != runtime.getClassLoader()) {
        // // Then replace context ClassLoader with AF one
        // Thread.currentThread().setContextClassLoader(runtime.getClassLoader());
        // }

        // Bind components in the assembly
        scaBindingProcess.processBinding(assembly, assemblyComponent,
            componentMap);

        // TODO Required with binding factory 0.5
        // restore context ClassLoader if necessary
        // if(Thread.currentThread().getContextClassLoader() !=
        // contextClassLoader)
        // Thread.currentThread().setContextClassLoader(contextClassLoader);
      }

    } catch (InstantiationException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    } catch (NoSuchInterfaceException e) {
      log.severe("Fractal Interface unavailable");
      e.printStackTrace();
    } catch (IllegalContentException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    } catch (IllegalLifeCycleException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }

    // return the assembly component
    return assemblyComponent;

  }

  /**
   * Create Fractal component type from the component definition in the model
   * 
   * @param comp
   *          - Component definition according to the model
   * @return - Resulting Component type
   * @throws InterfaceException
   */

  public ComponentType create_componentType(EObject comp)
      throws InterfaceException {

    try {

      // Create a new Collection of fractal interfaces type
      ArrayList<InterfaceType> interfaceType = new ArrayList<InterfaceType>();

      // Get list of services for this component/composite Object
      EList<?> slist = (EList<?>) comp.getClass().getMethod("getService")
          .invoke(comp);
      Iterator<?> si = slist.iterator();

      // Get list of refrences for this component/composite Object
      EList<?> rlist = (EList<?>) comp.getClass().getMethod("getReference")
          .invoke(comp);
      Iterator<?> ri = rlist.iterator();

      // Creates interface type for services
      while (si.hasNext()) {
        BaseService cs = (BaseService) si.next();
        try {
          interfaceType.add(itf.create(cs.getInterface()));
        } catch (InterfaceException e) {

          // Retrieve SCA Component
          EObject c = cs.eContainer();
          String name = (String) c.getClass().getMethod("getName").invoke(c);

          throw new InterfaceException("Could not create interface type for "
              + name + " with service " + cs.getName(), e);
        }
      }

      // Creates interface type for references
      while (ri.hasNext()) {
        BaseReference cr = (BaseReference) ri.next();
        try {
          interfaceType.add(itf.create(cr.getInterface()));
        } catch (InterfaceException e) {

          // Retrieve SCA Component
          EObject c = cr.eContainer();
          String name = (String) c.getClass().getMethod("getName").invoke(c);
          throw new InterfaceException("Could not create interface type for "
              + name + " with reference " + cr.getName(), e);
        }
      }

      InterfaceType[] itftype = interfaceType
          .toArray(new InterfaceType[interfaceType.size()]);

      log.finer("Create interface type for component");

      ComponentType componentType = runtime.createFcType(itftype);
      if (componentType == null)
        log.severe("Returned component type is NULL");

      return componentType;

    } catch (Exception e) {
      String name = null;
      try {
        name = (String) comp.getClass().getMethod("getName").invoke(comp);
      } catch (Exception e1) {
        // Nothing more to do
      }

      throw new InterfaceException(
          "Could not create interfaces type for component " + name, e);

    }

  }

  // --------------------------------------------------------------------------
  // Private methods
  // --------------------------------------------------------------------------

  /**
   * Initialize intents for a component, its services and its references
   * 
   * @param instanceComp
   *          the component upon which intents have to be set
   * @param modelComp
   *          representation of the component with required intents
   * @throws IntentException
   *           if an intent cannot be set
   */
  private void initIntents(org.objectweb.fractal.api.Component instanceComp,
      Component modelComp) throws IntentException {

    // Init requires for the component
    if (modelComp.getRequires() != null)
      intents.setIntent(instanceComp, modelComp);

    // Init requires for the component's references
    for (ComponentReference ref : modelComp.getReference()) {
      if (ref.getRequires() != null)
        intents.setIntent(instanceComp, ref);
    }

    // Init requires for the component's services
    for (ComponentService serv : modelComp.getService()) {
      if (serv.getRequires() != null)
        intents.setIntent(instanceComp, serv);
    }
  }

}
