/***
 * OW2 FraSCAti explorer
 * Copyright (C) 2008-2009 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org
 *
 * Author: Christophe Demarey
 *
 * Contributor(s):
 */
package org.ow2.frascati.explorer;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.fraclet.annotations.Controller;
import org.objectweb.fractal.fraclet.annotations.Interface;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.fraclet.types.Cardinality;
import org.objectweb.fractal.fraclet.types.Contingency;
import org.objectweb.util.explorer.api.Entry;
import org.objectweb.util.explorer.core.naming.lib.DefaultEntry;
import org.ow2.frascati.factory.runtime.domain.api.Domain;
import org.ow2.frascati.factory.runtime.domain.api.DomainConfig;

/**
 * This component is used to manage connections to SCA domains.
 * 
 * @author Christophe Demarey
 */
@org.objectweb.fractal.fraclet.annotations.Component(
    provides=@Interface(name="scadomain-manager", signature=DomainManager.class) )
public class DomainManagerImpl 
  implements DomainManager {

  // --------------------------------------------------------------------------
  // Internal state
  // --------------------------------------------------------------------------
  private static final long serialVersionUID = -7402725186973134104L;

  /** A reference to this component. */
  @Controller()
  protected Component self;

  /** The FraSCAti factory domain interfaces. */
  @Requires( name="domains", 
             contingency=Contingency.OPTIONAL,
             cardinality=Cardinality.COLLECTION )
  protected Map<String, Domain> domains = new HashMap<String, Domain>();
  
  /** The FraSCAti factory domain-config interfaces. */
  @Requires( name="configs",
             contingency=Contingency.OPTIONAL,
             cardinality=Cardinality.COLLECTION )
  protected Map<String, DomainConfig> domainConfigs = new HashMap<String, DomainConfig>();
  
  // --------------------------------------------------------------------------
  // Constructor
  // --------------------------------------------------------------------------

  /**
   * Defaul constructor
   */
  public DomainManagerImpl() { }
  
  // --------------------------------------------------------------------------
  // Private methods
  // --------------------------------------------------------------------------
  
  /**
   * Compute the domain name from the host name.
   * 
   * @param host - The name of the host providing the domain.
   * @return A unique identifier for the domain.
   */
  private String getDomainName(String host) {
    String name = "domain-" + host; 
    if ( domains.get(name) == null) return name;
   
    int i = 1;
    name += '-';
    while ( domains.get(name+i) != null)
      i++;
    return name+i;
  }

  // --------------------------------------------------------------------------
  // Public methods
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  // Methods for DomainManager
  // --------------------------------------------------------------------------
  
  /**
   * Initialize the domain manager.
   */
  public void init() {
    String baseDir = System.getProperty("java.io.tmpdir");
    
    // Set the output directory for generated files
    for (DomainConfig dc : domainConfigs.values()) {
      try {
        dc.setOutput( new File(baseDir) );
      } catch (IOException e) {
        e.printStackTrace();
      }
    }
  }
  
  /**
   * Get the list of SCA domains already bound to the Domain Manager.
   * 
   * @return a List of entry that can be used directly in the explorer tree.
   */
  public Entry[] getEntries(Object object) {
    Entry[] entries = new Entry[domains.size()];
    int i = 0;
    
    for (String name : domains.keySet()) {
      entries[i] = new DefaultEntry(name, domains.get(name));
      i++;
    }
    return entries;
  }
  
}
