/***
 * OW2 FraSCAti explorer
 * Copyright (C) 2008-2009 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org
 *
 * Author: Christophe Demarey
 *
 * Contributor(s):
 */
package org.ow2.frascati.explorer.context;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Interface;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.AttributeController;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.bf.connectors.rmi.RmiSkeletonAttributes;
import org.objectweb.fractal.bf.connectors.rmi.RmiStubAttributes;
import org.objectweb.fractal.bf.connectors.ws.WsSkeletonAttributes;
import org.objectweb.fractal.bf.connectors.ws.WsStubAttributes;
import org.objectweb.fractal.explorer.FcExplorer;
import org.objectweb.util.explorer.api.Context;
import org.objectweb.util.explorer.api.Entry;
import org.objectweb.util.explorer.core.naming.lib.DefaultEntry;
import org.ow2.frascati.tinfi.control.intent.IntentHandler;
import org.ow2.frascati.tinfi.control.intent.SCABasicIntentController;
import org.ow2.frascati.tinfi.control.intent.SCAIntentController;

/**
 * SCA interface context used to add SCA entries as childs of this node.
 *  
 * @author Christophe Demarey.
 */
public abstract class ScaInterfaceContext implements Context {
  /**
   * Tells if this interface has SCA intents or not.
   * 
   * @param owner The component owning this interface.
   * @param itfName The name of the interface to search intents on.
   * @return True if this interface has at least one intent.
   */
  private List<IntentHandler> getIntents(Component owner, String itfName) {
    SCABasicIntentController ic = null;

    try {
      ic = (SCABasicIntentController) owner.getFcInterface(SCAIntentController.NAME);
      return ic.listFcIntentHandler(itfName);
    } catch (NoSuchInterfaceException e) {
      return Collections.emptyList();
    }
  }

  /**
   * Get explorer tree entries for SCA intents.
   * 
   * @param owner The component owning this interface.
   * @param itfName The name of the interface to search intents on.
   * @return a List with an entry for each intent.
   */
  protected List<Entry> getIntentEntries(Component owner, String itfName) {
    List<Entry> entries = new ArrayList<Entry>();
    List<IntentHandler> intents = getIntents(owner, itfName);
    
    if ( !intents.isEmpty() ) {
      for (IntentHandler intent : intents) {
        Component intentComponent = ((Interface) intent).getFcItfOwner();
        entries.add( new DefaultEntry(FcExplorer.getName(intentComponent),
                     new IntentWrapper(intentComponent)) );
      }
    }
    return entries;
  }
  
  /**
   * Get explorer tree entries for SCA bindings.
   * 
   * @param owner The component owning this interface.
   * @param itfName The name of the interface to search bindings on.
   * @return a List with an entry for each binding.
   */
  protected List<Entry> getBindingEntries(Component owner, String itfName) {
    List<Entry> entries = new ArrayList<Entry>();
    Interface bindInterface = null;
    
    try {
      BindingController bc = FcExplorer.getBindingController(owner);
      bindInterface = (Interface) bc.lookupFc(itfName);
    } catch(Exception e) { 
      // Nothing to do.
    }
    if(bindInterface != null) {
      try {
        AttributeController ac = FcExplorer.getAttributeController( bindInterface.getFcItfOwner() );
        // WS binding
        if (   (ac instanceof WsSkeletonAttributes)
            || (ac instanceof WsStubAttributes) ) {
          entries.add( new DefaultEntry(FcExplorer.getPrefixedName(bindInterface),
              new WsBindingWrapper(bindInterface)) );
        }          
        // RMI binding
        if (   (ac instanceof RmiStubAttributes)
            || (ac instanceof RmiSkeletonAttributes) ) {
          entries.add( new DefaultEntry(FcExplorer.getPrefixedName(bindInterface),
              new RmiBindingWrapper(bindInterface)) );
        }          
      } catch (NoSuchInterfaceException e) {
        // Nothing to do
      }
    }
    return entries;
  }
  
}
