/***
 * OW2 FraSCAti Assembly Factory
 * Copyright (C) 2008-2009 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org 
 *
 * Author: Damien Fournier
 */

package org.ow2.frascati.factory.generate.runtime;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.util.Map;

import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.Type;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.fraclet.annotations.Component;
import org.objectweb.fractal.juliac.Juliac;
import org.objectweb.fractal.juliac.JuliacConfig;
import org.objectweb.fractal.juliac.plugin.CompileSupportException;
import org.objectweb.fractal.util.Fractal;
import org.ow2.frascati.factory.core.instance.runtime.RuntimeFactory;
import org.ow2.frascati.factory.core.instance.runtime.TinfiRuntime;

/**
 * GenericFactory and TypeFactory component service
 * 
 * @author <a href="mailto:damien.fournier@inria.fr">Damien Fournier</a>
 * @version $Revision$
 */

@Component
public class TinfiGenerate extends TinfiRuntime implements RuntimeGenerate,
    RuntimeFactory {

  /**
   * Logger for this class
   */

  java.util.logging.Logger log = java.util.logging.Logger
      .getLogger("assembly-factory");
  
  // Environment variable associated with default output directory
  protected static String FRASCATI_GENERATED = "FRASCATI_GENERATED";

  // Tinfi source code generator
  // TODO enable juliac/felix source code generator for OSGI support 
  protected static String JULIAC_OPT_LEVEL = "org.ow2.frascati.tinfi.juliac.FCOOCtrlSourceCodeGenerator"; // :org.objectweb.fractal.juliac.felix.FCOOCtrlSourceCodeGenerator";

  // Output Directory (Juliac BaseDir)
  protected File outputDir;

  // Output directory for generated sources
  protected static String GEN_DIRECTORY = "/target/generated-frascati-sources/";

  // Output directory for compiled classes
  // Must not be loaded by a ClassLoader which is not managed by the Assembly Factory
  protected static String CLASS_DIRECTORY = "/target/generated-frascati-classes/";

  protected Juliac jc;
  protected JuliacConfig jcfg;

  public TinfiGenerate() {

    super();

    try {

      jc = new Juliac();
      init();

      this.tf = Fractal.getTypeFactory(tinfi);

    } catch (NoSuchInterfaceException e) {
      e.printStackTrace();
    } catch (IllegalArgumentException e) {
      e.printStackTrace();
    } catch (IOException e) {
      e.printStackTrace();
    } 
  }

  /**
   * (non-Javadoc)
   * 
   * @see org.ow2.frascati.factory.generate.runtime.RuntimeGenerate#addJavaLib(java.lang.String)
   */

  public void addJavaLib(String[] lib) {
    jcfg.addSrclibs(lib);
  }

  /**
   * (non-Javadoc)
   * 
   * @see org.ow2.frascati.factory.generate.runtime.RuntimeGenerate#addJavaSource(java.lang.String)
   */

  public void addJavaSource(String[] source) {

    try {
      jcfg.addSrcs(source);
    } catch (IOException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }

  }

  /**
   * (non-Javadoc)
   * 
   * @see org.ow2.frascati.factory.generate.runtime.RuntimeGenerate#compile()
   */

  public void compile() throws IOException, IllegalArgumentException,
      CompileSupportException {
    jc.compile();
  }

  /**
   * (non-Javadoc)
   * 
   * @see org.ow2.frascati.factory.generate.runtime.RuntimeGenerate#init(java.io.File)
   */

  public void init() throws IOException {

    jcfg = new JuliacConfig(jc);
    jcfg.setOptLevel(JULIAC_OPT_LEVEL);
    jc.setJuliacConfig(jcfg);
    
    Map<String, String> map = System.getenv();
    String defaultOutputDir =  map.get(FRASCATI_GENERATED);
    
    File f = (defaultOutputDir == null) ? new File(".").getAbsoluteFile() : new File(defaultOutputDir);
    
    // Set default output dir
    setOutputDir(f);

  }

  /**
   * @see org.objectweb.fractal.api.factory.GenericFactory#newFcInstance(org.objectweb
   *      .fractal.api.Type, java.lang.Object, java.lang.Object)
   */

  // Invoke Juliac Compiler, does not return component instance
  @Override
  public org.objectweb.fractal.api.Component newFcInstance(Type arg0,
      Object arg1, Object arg2) throws InstantiationException {
    
    try {
      
      // test if an object array has been given for control desc
      if (arg1.getClass().isArray()) {
        // extract control desc information
        Object[] array = (Object[]) arg1;
        arg1 = array[1];
      }
      
      jc.generate(arg0, arg1, arg2);
      
    } catch (Exception e) {
      System.err
          .println("Error when trying to generate component for Class "
              + arg2);
      System.out.println("------ Debug ------");
      debug(jc.getJuliacConfig().getClassLoader());
      System.out.println("------ End debug ------");
      throw new Error(e);
    }
    return null;
  }

  /**
   * @see org.ow2.frascati.factory.instance.runtime.Loader#setClassLoader(java.lang.ClassLoader)
   */

  @Override
  public void setClassLoader(ClassLoader classloader) {
    super.setClassLoader(classloader);
    jcfg.setClassLoader(classloader);
    // this.currentClassloader = classloader;
  }

  public void setOutputDir(File outputDir) {
    if (jcfg == null)
      throw new Error("Must call init() first");

    if (!outputDir.equals(this.outputDir)) {
      try {
        this.outputDir = outputDir;
        jcfg.setBaseDir(outputDir);
        jcfg.setGenDirName(outputDir.getCanonicalPath() + GEN_DIRECTORY);
        jcfg.setClassDirName(outputDir.getCanonicalPath() + CLASS_DIRECTORY);

        // Ensure that target directory is referenced by current classloader
        // May be useless

        URL[] urls = { new File(outputDir.getCanonicalPath() + CLASS_DIRECTORY).toURI().toURL() };
        loadLibraries(urls);
        jcfg.setClassLoader(getClassLoader());
        
//        URLClassLoader cl = new URLClassLoader(urls, getClassLoader());
//        setClassLoader(cl);

      } catch (IOException e) {
        e.printStackTrace();
      }
    }

  }

  /** 
   * Debug output when having exception during component generation/instantiation
   * 
   * @see org.ow2.frascati.factory.core.instance.runtime.Loader#debug(java.lang.ClassLoader)
   */
  
  @Override
  public void debug(ClassLoader cl) {
    super.debug(cl);
    JuliacConfig jcfg = jc.getJuliacConfig();
    System.err.println("\nJuliac configuration details");

    String[] list;
    System.err.println("Base Directory  : " + jcfg.getBaseDir());
    System.err.println("Root Package    : " + jcfg.getPkgRoot());
    System.err.println("Directory for generated sources : "
        + jcfg.getGenDirName());
    System.err.println("Directory for compiled classes  : "
        + jcfg.getClassDirName());

    System.err.println("\n -----Loaded sources");
    list = jcfg.getSrcs();
    for (String src : list) {
      System.err.println(" " + src);
    }

    list = jcfg.getSrclibs();
    System.err.println("\n -----Loaded libraries");
    for (String lib : list) {
      System.err.println(" " + lib);
    }

  }

}
