/***
 * OW2 FraSCAti Tinfi
 * Copyright (C) 2007-2009 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.ow2.frascati.tinfi.juliac;

import java.io.IOException;
import java.util.List;
import java.util.Map;

import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.julia.type.BasicInterfaceType;
import org.objectweb.fractal.juliac.ClassGenerator;
import org.objectweb.fractal.juliac.ITBasedClassGenerator;
import org.objectweb.fractal.juliac.InitializerClassGenerator;
import org.objectweb.fractal.juliac.Juliac;
import org.objectweb.fractal.juliac.JuliacConfig;
import org.objectweb.fractal.juliac.JuliacLoader;
import org.objectweb.fractal.juliac.desc.ControllerDesc;
import org.objectweb.fractal.juliac.desc.MembraneDesc;
import org.objectweb.fractal.juliac.proxy.InterceptorSourceCodeGenerator;
import org.objectweb.fractal.juliac.proxy.InterfaceImplementationClassGenerator;
import org.objectweb.fractal.juliac.proxy.ProxyClassGenerator;
import org.objectweb.fractal.juliac.ucf.UnifiedClass;
import org.osoa.sca.annotations.Callback;

/**
 * This class generates the source code associated to Fractal components.
 * The membrane implementation and the initializer implementation are generated.
 * 
 * The content, the interceptors and the controllers are kept in separate
 * classes and the controllers are implemented with objects.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 */
public class FCOOCtrlSourceCodeGenerator
extends org.objectweb.fractal.juliac.opt.oo.FCOOCtrlSourceCodeGenerator {
    
    public FCOOCtrlSourceCodeGenerator() {
        super();
    }

    /** Configuration file for Tinfi. */
    public static final String DEFAULT_CONFIGURATION = "tinfi-bundled.cfg";

    
    // -----------------------------------------------------------------------
    // Implementation of the FCSourceCodeGeneratorItf interface
    // -----------------------------------------------------------------------

    /**
     * Initialize this generator.
     * 
     * The initialization process is kept separated from the constructor in
     * order to be able to customize the julia.config system property.
     * 
     * @param jc  the instance of Juliac using this source code generator
     */
    @Override
    public void init( Juliac jc ) throws IOException {

        JuliacLoader loader = jc.getJuliacConfig().getJuliacLoader();
        loader.loadConfigFile(DEFAULT_CONFIGURATION);
        
        super.init(jc);

        /*
         * Register component interceptor code generators.
         */
        JuliacConfig jconf = jc.getJuliacConfig();
        jconf.putInterceptorSourceCodeGenerator(
            "org.ow2.frascati.tinfi.asm.SCAContentInterceptorCodeGenerator",
            SCAContentInterceptorSourceCodeGenerator.class.getName() );
        jconf.putInterceptorSourceCodeGenerator(
            "org.ow2.frascati.tinfi.asm.SCAIntentInterceptorCodeGenerator",
            SCAIntentInterceptorSourceCodeGenerator.class.getName() );
    }

    /**
     * Return <code>true</code> if the current source code generator handles the
     * specified controller descriptor.
     * 
     * @param controllerDesc  the controller descriptor to be checked
     * @return                if <code>controllerDesc</code> should be included 
     * @since 2.2
     */
    @Override
    public boolean accept( Object controllerDesc ) {
        boolean accept =
            controllerDesc.equals("scaPrimitive") ||
            controllerDesc.equals("scaComposite");
        return accept;
    }
    
    /**
     * Return the source code generators for the initializer classes associated
     * with this component source code generator.
     */
    @Override
    public InitializerClassGenerator<List<ControllerDesc>> getInitializerClassGenerator() {
        return new InitializerOOCtrlClassGenerator();
    }

    /**
     * Return the source code generator for component interfaces.
     */
    @Override
    public ProxyClassGenerator getInterfaceClassGenerator( 
        InterfaceType it, MembraneDesc<?> membraneDesc ) {

        String ctrlDesc = membraneDesc.getDescriptor();
        InterfaceSourceCodeGeneratorFactory iscgf =
            new InterfaceSourceCodeGeneratorFactory(it,ctrlDesc,membraneDesc,jc);
        ProxyClassGenerator pcg = iscgf.getInterfaceSourceCodeGenerator();
        return pcg;
    }
    
    /**
     * Generate the source code of component interfaces associated with the
     * specified interface type.
     */
    @Override
    protected void generateInterfaceImpl( InterfaceType it, String ctrlDesc )
    throws IOException {
        
        InterfaceSourceCodeGeneratorFactory iscgf =
            new InterfaceSourceCodeGeneratorFactory(it,ctrlDesc,null,jc);
        
        if( iscgf.isSCACompBusinessInterface() ) {
           
            /*
             * Generate SCA specific interface implementations.
             */

            // Server (input) interfaces
            ClassGenerator cg = iscgf.getInterfaceSourceCodeGenerator();
            jc.generateSourceCode(cg);
                
            // ServiceReference implementation
            cg = new ServiceReferenceClassGenerator(it,null,false,jc);
            jc.generateSourceCode(cg);
            
            // CallableReference and input interface implementations for callbacks
            String signature = it.getFcItfSignature();
            UnifiedClass cl = jc.create(signature);
            Callback annot = cl.getAnnotation(Callback.class);
            if( annot != null ) {
                Class<?> cbcl = annot.value();
                InterfaceType cbit =
                    new BasicInterfaceType(
                        "callback", cbcl.getName(), false, false, false );
                cg = new CallableReferenceClassGenerator(cbit,null,false,jc);
                jc.generateSourceCode(cg);
                
                cg = new CallBackInterfaceClassGenerator(cbit,null,false,jc);
                jc.generateSourceCode(cg);
            }
        }
        else {
            
            /*
             * Interfaces of non-SCA components.
             */
            ClassGenerator cg = 
                new InterfaceImplementationClassGenerator(it,null,false,jc);
            jc.generateSourceCode(cg);
        }
    }

    /**
     * Return the interceptor class generator for the specified controller
     * descriptor. Return <code>null</code> if the specified controller
     * descriptor is not associated with an interceptor source code generator. 
     */
    @Override
    protected ITBasedClassGenerator getInterceptorClassGenerator(
        String ctrldesc, Map<?,?> context )
    throws ClassNotFoundException {
        
        if( ! ctrldesc.startsWith("sca") ) {
            /*
             * For non SCA controller descriptors, use the regular
             * InterceptorClassGenerator class defined by the super method.
             */
            ITBasedClassGenerator itcg =
                super.getInterceptorClassGenerator(ctrldesc,context);
            return itcg;
        }
        
        // Get the interceptor source code generators from the Julia config file
        boolean mergeable = false;
        InterceptorSourceCodeGenerator[] scgs =
            getInterceptorSourceCodeGenerator(ctrldesc,context,mergeable);
        if( scgs == null ) {
            // No interceptor
            return null;
        }
        
        // InterceptorClassGenerator specific for SCA components
        InterceptorClassGenerator icg = new InterceptorClassGenerator(scgs,jc);
        return icg;
    }    
}
