/***
 * OW2 FraSCAti Tinfi
 * Copyright (C) 2007-2009 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.ow2.frascati.tinfi;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Stack;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Interface;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.julia.InitializationContext;
import org.objectweb.fractal.julia.Interceptor;
import org.osoa.sca.RequestContext;
import org.ow2.frascati.tinfi.control.component.ReconfigurableComponentContext;
import org.ow2.frascati.tinfi.control.content.SCAContentController;
import org.ow2.frascati.tinfi.control.intent.IntentHandler;

/**
 * The root class for Tinfi component interceptors.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 */
public abstract class TinfiComponentInterceptor implements Interceptor {

    // -----------------------------------------------------------------------
    // Partial implementation of the Interceptor (and Controller) interfaces
    // Abstract methods are concretized by generated subclasses
    // -----------------------------------------------------------------------
    
    public void initFcController( InitializationContext ic )
    throws InstantiationException {
        
        Interceptor owner = (Interceptor) ic.getInterface("component");
        
        /*
         * Skip the intent interceptor which is installed on all control
         * interfaces, including "component".
         */
        this.owner = (Component) owner.getFcItfDelegate();
    }
    
    @Override
    public abstract Object clone();
    
    /**
     * Initialize a clone instance. This method is called by generated
     * subclasses.
     */
    protected void initFcClone( TinfiComponentInterceptor clone ) {
        clone.intentHandlersMap = intentHandlersMap;
        clone.owner = owner;
        clone.itf = itf;
    }

    private Component owner;
    private ReconfigurableComponentContext compctx;
    private SCAContentController cc;
    private Interface itf;

    
    // -----------------------------------------------------------------------
    // Methods for managing the reference between the interceptor and its
    // associated interface
    // -----------------------------------------------------------------------

    public void setFcItf( Interface itf ) {
        this.itf = itf;
    }
    
    public Interface getFcItf() {
        return itf;
    }
    
    
    // -----------------------------------------------------------------------
    // Management of intents handlers
    // -----------------------------------------------------------------------
    
    /*
     * The Fc string in method names has been removed to avoid name clashes when
     * adding intent handlers on the SCAIntentController interface.
     */
    
    /**
     * The map of intent handlers associated with each component method managed
     * by this interceptor. This map is queried by generated subclasses.
     * 
     * @since 0.4
     */
    protected Map<Method,List<IntentHandler>> intentHandlersMap;
    
    /**
     * The array of component methods managed by this interceptor.
     * 
     * @since 0.6
     */
    protected Method[] methods;
        
    /**
     * Initialize the intent handlers map by adding an empty entry for each
     * specified method. 
     */
    protected void initIntentHandlersMap( Method[] methods ) {
        this.methods = methods;
        intentHandlersMap = new HashMap<Method,List<IntentHandler>>();
        for (Method method : methods) {
            List<IntentHandler> handlers = new ArrayList<IntentHandler>();
            intentHandlersMap.put(method,handlers);
        }
    }
    
    /**
     * Add the specified intent handler for all component methods managed by
     * this interceptor.
     */
    public void addIntentHandler( IntentHandler handler ) {
        for (Map.Entry<Method,List<IntentHandler>> entry : intentHandlersMap.entrySet()) {
            List<IntentHandler> handlers = entry.getValue();
            handlers.add(handler);
        }
    }
    
    /**
     * Add the specified intent handler for the specified method.
     * 
     * @throws NoSuchMethodException
     *      if the specified method is not a component method managed by this
     *      interceptor
     */
    public void addIntentHandler( IntentHandler handler, Method method )
    throws NoSuchMethodException {
        
        if( ! intentHandlersMap.containsKey(method) ) {
            throw new NoSuchMethodException();
        }
        
        List<IntentHandler> handlers = intentHandlersMap.get(method);
        handlers.add(handler);
    }
    
    /**
     * Return the array of component methods managed by this interceptor.
     * 
     * @since 0.6
     */
    public Method[] getMethods() {
        return methods;
    }
    
    /**
     * Return the list of all intent handlers managed by this interceptor.
     * 
     * @since 0.4.4
     */
    public List<IntentHandler> listIntentHandler() {
        List<IntentHandler> all = new ArrayList<IntentHandler>();
        for (Map.Entry<Method,List<IntentHandler>> entry : intentHandlersMap.entrySet()) {
            List<IntentHandler> handlers = entry.getValue();
            all.addAll(handlers);
        }
        return all;
    }
    
    /**
     * Return the list of all intent handlers managed by this interceptor for
     * the specified method.
     * 
     * @throws NoSuchMethodException
     *      if the specified method is not a component method managed by this
     *      interceptor
     * @since 0.4.4
     */
    public List<IntentHandler> listIntentHandler( Method method )
    throws NoSuchMethodException {

        if( ! intentHandlersMap.containsKey(method) ) {
            throw new NoSuchMethodException();
        }
        
        List<IntentHandler> handlers = intentHandlersMap.get(method);
        List<IntentHandler> all = new ArrayList<IntentHandler>();
        all.addAll(handlers);
        return all;
    }
    
    /**
     * Remove the specified intent handler for all component methods managed by
     * this interceptor.
     */
    public void removeIntentHandler( IntentHandler handler ) {
        for (Map.Entry<Method,List<IntentHandler>> entry : intentHandlersMap.entrySet()) {
            List<IntentHandler> handlers = entry.getValue();
            handlers.remove(handler);
        }
    }
    
    /**
     * Remove the specified intent handler for the specified method.
     * 
     * @throws NoSuchMethodException
     *      if the specified method is not a component method managed by this
     *      interceptor
     */
    public void removeIntentHandler( IntentHandler handler, Method method )
    throws NoSuchMethodException {
        
        if( ! intentHandlersMap.containsKey(method) ) {
            throw new NoSuchMethodException();
        }
        
        List<IntentHandler> handlers = intentHandlersMap.get(method);
        handlers.remove(handler);
    }
    
    
    // -----------------------------------------------------------------------
    // Implementation specific
    // -----------------------------------------------------------------------
    
    /**
     * Return the thread-local stack of request contexts.
     */
    protected Stack<RequestContext> getFcRequestContextStack() {
        SCAContentController cc = getFcSCAContentCtrlItf();
        Stack<RequestContext> stack = cc.getRequestContextStack();
        if( stack == null ) {
          stack = new Stack<RequestContext>();
          cc.setRequestContextStack(stack);
        }
        return stack;
    }

    /*
     * The Fc string in this method name has been removed to avoid name clashes
     * when adding intent handlers on the SCAContentController interface which
     * defines a getFcContent method.
     */
    
    /**
     * Query the content controller to retrieve the content instance
     * corresponding to the current scope policy.
     */
    protected Object getContent() {
        SCAContentController cc = getFcSCAContentCtrlItf();
        try {
            Object content = cc.getFcContent();
            return content;
        }
        catch( InstantiationException e ) {
            throw new TinfiRuntimeException(e);
        }
    }
    
    /**
     * Release the specified content instance previously retrieved with {@link
     * #getFcContent()}.
     * 
     * @param content  the content instance which is released
     * @param isEndMethod
     *      true if the method which releases the content instance is annotated
     *      with {@link org.osoa.sca.annotations.EndsConversation}.
     */
    protected void releaseFcContent( Object content, boolean isEndMethod ) {
        SCAContentController cc = getFcSCAContentCtrlItf();
        cc.releaseFcContent(content,isEndMethod);
    }
    
    /**
     * Return the reference of the component context control interface
     * associated with the current component. This method is used by generated
     * subclasses.
     */
    protected ReconfigurableComponentContext getFcCompCtxCtrlItf() {
        if( compctx == null ) {
            compctx =
                getFcInterface(
                    ReconfigurableComponentContext.NAME,
                    ReconfigurableComponentContext.class );
        }
        return compctx;
    }
    
    /**
     * Return the reference of the SCA content control interface associated with
     * the current component.
     */
    private SCAContentController getFcSCAContentCtrlItf() {
        if( cc == null ) {
            cc = getFcInterface(SCAContentController.NAME,SCAContentController.class);
        }
        return cc;
    }
    
    /**
     * Return the interface of the specified name and type associated with the
     * current component.
     * 
     * @param name  the interface name
     * @param type  the interface type
     * @throws TinfiRuntimeException  if the interface does not exist
     */
    private <T> T getFcInterface( String name, Class<T> type )
    throws TinfiRuntimeException {
        try {
            T itf = (T) owner.getFcInterface(name);
            return itf;
        }
        catch( NoSuchInterfaceException e ) {
            throw new TinfiRuntimeException(e);
        }
    }
}
