/***
 * OW2 FraSCAti Tinfi
 * Copyright (C) 2007-2009 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.ow2.frascati.tinfi.control.content;

import java.lang.annotation.Annotation;
import java.lang.reflect.AccessibleObject;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.ow2.frascati.tinfi.TinfiRuntimeException;
import org.ow2.frascati.tinfi.metadata.MetaData;

/**
 * This class manages {@link Field}-based or property-based (getter/setter)
 * {@link Method} elements.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 */
public class AnnotatedElementsManager {
    
    /** The annotated elements. */
    private Map<String,AnnotatedElement> links =
        new HashMap<String,AnnotatedElement>();
    
    /** The instance for accessing the class and its metadata. */
    private MetaData<?> cmd;
    
    /** The annotation. */
    private Class<? extends Annotation> annot;
    
    public AnnotatedElementsManager(
        MetaData<?> cmd, Class<? extends Annotation> annot ) {
        
        this.cmd = cmd;
        this.annot = annot;
    }
    
    /**
     * Register all annotated elements.
     * 
     * @throws AnnotatedElementNotFoundException   
     *      if the reference does not exist
     * @throws DuplicateAnnotatedElementException
     *      if the reference is declared more than once
     */
    public void addAllReferences()
    throws
        AnnotatedElementNotFoundException, DuplicateAnnotatedElementException {        
        
        AccessibleObject[] aos = cmd.getAllAnnotatedMethodsAndFields(annot);
        
        Set<String> names = new HashSet<String>();
        for (AccessibleObject ao : aos) {
            
            /*
             * Check whether the AccessibleObject has a 'name' annotation
             * parameter.
             */
            Annotation a = ao.getAnnotation(annot);
            Class<?> acl = a.getClass();
            try {
                Method m = acl.getMethod("name");
                try {
                    String name = (String) m.invoke(a);
                    if( name.length() != 0 ) {
                        names.add(name);
                        continue;
                    }
                }
                catch (Exception e) {
                    throw new TinfiRuntimeException(e);
                }
            }
            catch (NoSuchMethodException e) {}
            
            /*
             * No 'name' annotation parameter found.
             * Infer the name of the dependency from the setter method name or
             * the field property.
             */
            String name = null;
            if( ao instanceof Method ) {
                Method method = (Method) ao;
                String s = method.getName();
                name = s.substring(3);   // skip 'set'
                name = name.substring(0,1).toLowerCase()+name.substring(1);
            }
            else {
                Field field = (Field) ao;
                name = field.getName();
            }
            names.add(name);
        }
        
        for (String name : names) {
            addReference(aos,name);
        }
    }
    
    /**
     * Register an annotated element with the specified name.
     * 
     * @param name  the name of the annotated element
     * 
     * @throws AnnotatedElementNotFoundException
     *      if the reference does not exist
     * @throws DuplicateAnnotatedElementException
     *      if the reference is declared more than once
     */
    public void addReference( String name )
    throws
        AnnotatedElementNotFoundException, DuplicateAnnotatedElementException {        
        
        AccessibleObject[] aos = cmd.getAllAnnotatedMethodsAndFields(annot);
        addReference(aos,name);
    }
    
    /**
     * Return <code>true</code> if the set of managed reference names contains
     * the specified name.
     */
    public boolean containsReferenceName( String name ) {
        return links.containsKey(name);
    }
    
    /**
     * Return the set of managed reference names.
     */
    public Set<String> getReferenceNames() {
        /*
         * Return a copy to let it be manipulated independently from the
         * elements stored in {@link #links}.
         */
        return new HashSet<String>( links.keySet() );
    }
    
    /**
     * Return the annotation associated with the specified name.
     * 
     * @param name    the name of the reference
     * @return        the annotation associated to the reference
     * @throws AnnotatedElementNotFoundException
     *      if the reference does not exist
     */
    public Annotation getAnnotation( String name )
    throws AnnotatedElementNotFoundException {
        
        AnnotatedElement re = links.get(name);
        if( re == null ) {
            throw new AnnotatedElementNotFoundException(
                        name, cmd.getSupportClass(), annot );
        }
        
        Annotation ref = re.getAnnotation(annot);
        return ref;
    }

    /**
     * Set the value of the reference in the target object.
     * 
     * @param target  the target object
     * @param name    the name of the reference
     * @param value   the specified value
     * 
     * @throws AnnotatedElementNotFoundException
     *      if the reference does not exist
     * @throws IllegalAnnotatedElementOperationException
     *      in case of problem when accessing the reference
     */
    public void set( Object target, String name, Object value )
    throws
        AnnotatedElementNotFoundException,
        IllegalAnnotatedElementOperationException {
        
        AnnotatedElement re = links.get(name);
        if( re == null ) {
            throw new AnnotatedElementNotFoundException(
                        name, cmd.getSupportClass(), annot );
        }
        
        try {
            re.set( target, value );
        }
        catch( IllegalAccessException e ) {
            throw new IllegalAnnotatedElementOperationException(e);
        }
        catch( InvocationTargetException e ) {
            throw new IllegalAnnotatedElementOperationException(e);
        }
    }

    /**
     * Return the value of the reference in the target object.
     * 
     * @param target  the target object
     * @param name    the name of the reference
     * @return        the value of the reference
     * 
     * @throws AnnotatedElementNotFoundException
     *      if the reference does not exist
     * @throws IllegalAnnotatedElementOperationException
     *      in case of problem when accessing the reference
     */
    public Object get( Object target, String name )
    throws
        AnnotatedElementNotFoundException,
        IllegalAnnotatedElementOperationException {
        
        AnnotatedElement re = links.get(name);
        if( re == null ) {
            throw new AnnotatedElementNotFoundException(
                        name, cmd.getSupportClass(), annot );
        }

        try {
            Object value = re.get(target);
            return value;
        }
        catch( IllegalAccessException e ) {
            throw new IllegalAnnotatedElementOperationException(e);
        }
    }

    
    // ---------------------------------------------------------------------
    // Implementation specific
    // ---------------------------------------------------------------------
    
    /**
     * Register an annotated element for the specified name selected among the
     * specified array of {@link AccessibleObject}s.
     * 
     * @param aos0  the {@link AccessibleObject}s
     * @param name  the name of the reference
     * 
     * @throws AnnotatedElementNotFoundException
     *      if the reference does not exist
     * @throws DuplicateAnnotatedElementException
     *      if the reference is declared more than once
     */
    private void addReference( AccessibleObject[] aos0, String name )
    throws
        AnnotatedElementNotFoundException, DuplicateAnnotatedElementException {
        
        AccessibleObject[] aos = getAllAnnotatedFieldsAndMethodsFor(aos0,name);
        
        if( aos.length < 1 ) {
            throw new AnnotatedElementNotFoundException(
                        name, cmd.getSupportClass(), annot );
        }
        if( aos.length > 1 ) {
            throw new DuplicateAnnotatedElementException(
                        name, cmd.getSupportClass(), annot );
        }
        
        if( aos[0] instanceof Method ) {
            Method method = (Method) aos[0];
            AnnotatedElement re = new AnnotatedElement(method);
            links.put(name,re);
        }
        else {
            Field field = (Field) aos[0];
            AnnotatedElement re = new AnnotatedElement(field);
            links.put(name,re);
        }
    }
    
    /**
     * Select and return in the specified {@link AccessibleObject}s, the ones
     * which correspond to a dependency to the specified reference name.
     * 
     * @param oas      an array of {@link AccessibleObject}s
     * @param refname  a reference name
     * @return
     *      the {@link AccessibleObject}s from <code>oas</code> which correspond
     *      to a dependency to <code>refname</code>
     */
    private AccessibleObject[] getAllAnnotatedFieldsAndMethodsFor(
        AccessibleObject[] aos, String refname ) {
        
        List<AccessibleObject> result = new ArrayList<AccessibleObject>();
        for (AccessibleObject ao : aos) {
            
            if( ao instanceof Constructor<?> ) {
                continue;
            }
            
            if( nameAnnotationParameterEquals(ao,refname) ) {
                result.add(ao);
            }
            else if( ao instanceof Method ) {
                Method method = (Method) ao;
                try {
                    Util.checkSetterMethodSignature(method,refname,null);
                    result.add(ao);
                }
                catch( IllegalArgumentException iae ) {}
            }
            else {
                // (ao instanceof Field) == true
                Field field = (Field) ao;
                if( field.getName().equals(refname) ) {
                    result.add(ao);
                }
            }
        }
        
        return result.toArray(new AccessibleObject[result.size()]);
    }
    
    /**
     * Return <code>true</code> if the specified accessible object is annotated
     * and if the annotation declares a 'name' parameter with a value equals to
     * the given value.
     */
    private boolean nameAnnotationParameterEquals(
        AccessibleObject ao, String value ) {
        
        try {
            Method meth = annot.getMethod("name");
            Annotation ref = ao.getAnnotation(annot);
            try {
                String v = (String) meth.invoke(ref);
                if( v!=null && v.equals(value) ) {
                    return true;
                }
            }
            catch( Exception e ) {
                throw new TinfiRuntimeException(e);
            }
        }
        catch( NoSuchMethodException e ) {}
        
        return false;
    }
}
