/***
 * OW2 FraSCAti Tinfi
 * Copyright (C) 2008-2009 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org
 *
 * Author: Lionel Seinturier
 * Contributor: Philippe Merle
 */

package org.ow2.frascati.tinfi.control.content.scope;

import java.lang.annotation.Annotation;
import java.lang.reflect.AccessibleObject;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Interface;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.NameController;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.julia.ComponentInterface;
import org.objectweb.fractal.julia.Interceptor;
import org.objectweb.fractal.julia.factory.ChainedInstantiationException;
import org.osoa.sca.CallableReference;
import org.osoa.sca.RequestContext;
import org.osoa.sca.ServiceReference;
import org.osoa.sca.annotations.Constructor;
import org.osoa.sca.annotations.Property;
import org.osoa.sca.annotations.Reference;
import org.ow2.frascati.tinfi.CallbackManager;
import org.ow2.frascati.tinfi.TinfiComponentOutInterface;
import org.ow2.frascati.tinfi.TinfiRuntimeException;
import org.ow2.frascati.tinfi.control.component.ReconfigurableComponentContext;
import org.ow2.frascati.tinfi.control.content.AnnotatedElementNotFoundException;
import org.ow2.frascati.tinfi.control.content.ContentClassMetaData;
import org.ow2.frascati.tinfi.control.content.IllegalAnnotatedElementOperationException;
import org.ow2.frascati.tinfi.control.property.SCAPropertyController;

/**
 * Root class for scope managers.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @author Philippe Merle <Philippe.Merle@inria.fr>
 */
public class AbstractScopeManager {
    
    protected Component comp;
    protected SCAPropertyController scapc;
    protected ReconfigurableComponentContext compCtx;
    protected NameController nc;
    
    /**
     * The metadata for the content class associated to the current
     * scoped-component.
     */
    protected ContentClassMetaData ccmd;
    
    public AbstractScopeManager(
        Component comp, SCAPropertyController scapc,
        ReconfigurableComponentContext compCtx, NameController nc,
        ContentClassMetaData ccmd ) {
        
        this.comp = comp;
        this.scapc = scapc;
        this.compCtx = compCtx;
        this.nc = nc;
        this.ccmd = ccmd;
    }

    
    /**
     * Create and initialize a content instance.
     */
    protected Object createAndInitializeFcInstance()
    throws InstantiationException {
        
        List<String> initnames = new ArrayList<String>();
        
        try {
            Object content = createFcInstance(initnames);
            initializeFcInstance(content,initnames);
            return content;
        }
        catch( Exception e ) {
            if( e instanceof InstantiationException ) {
                throw (InstantiationException) e;
            }
            if( e instanceof RuntimeException ) {
                throw (RuntimeException) e;
            }
            throw new ChainedInstantiationException(e,comp,e.getMessage());
        }
    }
    
    /**
     * Create a new content instance.
     * 
     * @param initnames
     *      this list is empty when the method is called and populated with the
     *      reference and property names initialized while creating the content
     *      instance with an @{@link Constructor} annotated constructor
     */
    private Object createFcInstance( List<String> initnames )
    throws
        java.lang.InstantiationException, IllegalAccessException,
        InstantiationException, InvocationTargetException {
        
        Object content = null;
        
        if( ccmd.constructorAnnotatedElement == null ) {
            content = ccmd.fcContentClass.newInstance();
        }
        else {
            ComponentType ct = (ComponentType) comp.getFcType();

            org.osoa.sca.annotations.Constructor annot =
                ccmd.constructorAnnotatedElement.getAnnotation(
                    org.osoa.sca.annotations.Constructor.class );
            
            /*
             * @Constructor defines a String[] value property which defaults to
             * [""]. Detect this default value.
             * 
             * Potential issue: if the value is really [""]?
             */
            String[] annotvalue = annot.value();
            if( annotvalue.length>1 || annotvalue[0].length()>0 ) {
                for (String av : annotvalue) {
                    initnames.add(av);
                }
            }
                
            /*
             * Extra property names may be specified by annotating the
             * parameters of the @Constructor annotated contructor.
             */
            Annotation[][] paramsAnnots =
                ccmd.constructorAnnotatedElement.getParameterAnnotations();
            for (Annotation[] paramAnnots : paramsAnnots) {
                for (Annotation pannot : paramAnnots) {
                    if( pannot instanceof Property ) {
                        Property propannot = (Property) pannot;
                        String name = propannot.name();
                        initnames.add(name);
                    }
                }
            }
            
            Object[] values = new Object[initnames.size()];
            for (int i = 0 ; i < initnames.size() ; i++) {
                String name = initnames.get(i);
                try {
                    InterfaceType it = ct.getFcInterfaceType(name);
                    Object sr = getServiceReference(it);
                    values[i] = sr;
                }
                catch( NoSuchInterfaceException nsie ) {
                    /*
                     * Not a reference. Check whether this is a property.
                     */
                    if( ! scapc.containsKey(name) ) {
                        /*
                         * Not a reference, nor a property.
                         */
                        String msg =
                            "Identifier "+name+
                            " in @Constructor attributes for class "+
                            ccmd.fcContentClass+
                            " is neither a reference name, nor a property name";
                        throw new InstantiationException(msg);
                    }
                    values[i] = scapc.get(name);
                }
            }
            
            /*
             * Construct the content instance by invoking the constructor with
             * initialization values.
             */
            content = ccmd.constructorAnnotatedElement.newInstance(values);
        }
        
        return content;
    }
    
    /**
     * Initialize newly created content instances.
     * 
     * @param content  the content instance
     * @param initnames
     *      the reference and property names which has been already initialized
     *      on the specified content instance
     */
    private void initializeFcInstance( Object content, List<String> initnames )
    throws
        InstantiationException, AnnotatedElementNotFoundException,
        IllegalAccessException, InvocationTargetException,
        IllegalAnnotatedElementOperationException, NoSuchInterfaceException {
        
        /*
         * Set the component context.
         */
        if( ccmd.contextAnnotatedElement != null ) {
            ccmd.contextAnnotatedElement.set(content,compCtx);
        }
        
        /*
         * Set the component name.
         */
        if( ccmd.componentNameAnnotatedElement != null ) {
            String name = nc.getFcName();
            ccmd.componentNameAnnotatedElement.set(content,name);
        }
        
        /*
         * Set references.
         * 
         * I was previously interating on ccmd.refs.getReferenceNames() for
         * injecting references. However POJO setters are not included in this
         * set (it would be a bit difficult to decide whether they should be
         * included in ccmd.refs or in ccmd.props).
         */
        ComponentType ct = (ComponentType) comp.getFcType();
        InterfaceType[] its = ct.getFcInterfaceTypes();
        for (InterfaceType it : its) {
            
            String name = it.getFcItfName();
            if( name.equals("component") || name.endsWith("-controller") ) {
                // Skip control interfaces
                continue;
            }
            if( ! it.isFcClientItf() ) {
                // Skip server interfaces
                continue;
            }
            
            Object sr = getServiceReference(it);
            if( ccmd.refs.containsReferenceName(name) ) {
                // @Reference annotation
                ccmd.refs.set(content,name,sr);     // Inject the reference
                if( sr instanceof List ) {
                    // Collection references
                    List<?> l = (List<?>) sr;
                    for (Object o : l) {
                        checkMandatoryReferenceHasBeenSet(o,name);
                    }
                }
                else {
                    // Singleton references
                    checkMandatoryReferenceHasBeenSet(sr,name);
                }
            }
            else {
                if( ccmd.setters.containsKey(name) ) {
                    // POJO
                    Method setter = ccmd.setters.get(name);
                    setter.invoke(content,sr);
                }
                else {
                    /*
                     * Indeed nothing.
                     * 
                     * Neither an annotated case, nor a POJO. The reference may
                     * have been injected via an @Constructor annotation. See
                     * {@link #createFcInstance()}.
                     */
                    if( ! initnames.contains(name) ) {
                        String msg =
                            "Reference or property "+name+
                            " can not be injected on "+ccmd.fcContentClass;
                        throw new InstantiationException(msg);
                    }
                }                
            }
        }

        /*
         * Set properties.
         * 
         * I was previously interating on ccmd.props.getReferenceNames() for
         * injecting property values. However POJO setters are not included in
         * this set (it would be a bit difficult to decide whether they should
         * be included in ccmd.refs or in ccmd.props).
         */
        Map<String,Object> props = scapc.getAll();
        for (Map.Entry<String,Object> entry : props.entrySet()) {
            String name = entry.getKey();
            Object value = entry.getValue();
            if( ccmd.props.containsReferenceName(name) ) {
                // @Property annotation
                ccmd.props.set(content,name,value);
            }
            else {
                if( ccmd.setters.containsKey(name) ) {
                    // POJO
                    Method setter = ccmd.setters.get(name);
                    setter.invoke(content,value);                
                }
                else {
                    /*
                     * Indeed nothing.
                     * 
                     * Neither an annotated case, nor a POJO. The property value
                     * may have been injected via an @Constructor annotation.
                     * See {@link #createFcInstance()}.
                     */
                    if( ! initnames.contains(name) ) {
                        String msg =
                            "Reference or property "+name+
                            " can not be injected on "+ccmd.fcContentClass;
                        throw new InstantiationException(msg);
                    }
                }
            }
        }
        
        // Check that all mandatory properties have been set
        Set<String> names = ccmd.props.getReferenceNames();
        for (String name : names) {
            if( ! scapc.containsKey(name) ) {
                Property prop = (Property) ccmd.props.getAnnotation(name);
                if( prop.required() ) {
                    String msg =
                        "Mandatory property(s) "+name+" in component "+
                        nc.getFcName()+" should have been set";
                    throw new InstantiationException(msg);
                }
            }
        }
        
        /*
         * Set callbacks.
         */
        if( ccmd.callbacks.length > 0 ) {
            Stack<CallableReference<?>> stack = CallbackManager.get();
            if( ! stack.isEmpty() ) {
                CallableReference<?> cr = stack.peek();
                Class<?> businessItf = cr.getBusinessInterface();
                
                for (AccessibleObject cb : ccmd.callbacks) {
                    if( cb instanceof Method ) {
                        Method method = (Method) cb;
                        Class<?>[] params = method.getParameterTypes();
                        if( params[0].isAssignableFrom(businessItf) ) {
                            // initFcController checked that this was a setter
                            method.invoke(content,new Object[]{cr.getService()});
                            break;
                        }
                    }
                    else {
                        Field field = (Field) cb;
                        Class<?> type = field.getType();
                        if( type.isAssignableFrom(businessItf) ) {
                            field.set(content,cr.getService());
                            break;
                        }
                    }
                }            
            }
        }
        
        /*
         * Invoke the @Init annotated method, if any.
         */
        if( ccmd.initMethod != null ) {
            ccmd.initMethod.invoke(content,new Object[]{});
        }
    }
    
    /**
     * If the specified reference is mandatory, check that it has been set.
     * 
     * @param sr    the service reference
     * @param name  the name of the service reference
     * @throws AnnotatedElementNotFoundException
     *      if the reference can not be found
     * @throws InstantiationException
     *      if the reference has not been set
     */
    private void checkMandatoryReferenceHasBeenSet( Object sr, String name )
    throws AnnotatedElementNotFoundException, InstantiationException {
        
        ServiceReference<?> sref = (ServiceReference<?>) sr;
        ComponentInterface ci = (ComponentInterface) sref.getService();
        Interceptor intercept = (Interceptor) ci.getFcItfImpl();  // Skip the interceptor
        Object impl = intercept.getFcItfDelegate();
        if( impl == null ) {
            Reference ref = (Reference) ccmd.refs.getAnnotation(name);
            if (ref.required()) {
                String msg =
                    "Mandatory reference " + name + " in component " +
                    nc.getFcName() + " should have been set";
                throw new InstantiationException(msg);
            } 
        }        
    }
    
    /**
     * Return the {@link ServiceRefence} (or the list of {@link
     * ServiceReference}s) associated to the specified {@link InterfaceType}.
     */
    private Object getServiceReference( InterfaceType it )
    throws NoSuchInterfaceException {
        
        String name = it.getFcItfName();
        if( it.isFcCollectionItf() ) {
            Object[] itfs = comp.getFcInterfaces();
            List<Object> srs = new ArrayList<Object>();
            for (Object o : itfs) {
                Interface itf = (Interface) o;
                String itfname = itf.getFcItfName();
                if( itfname.startsWith(name) ) {
                    Object sr = getServiceReference(itf);
                    srs.add(sr);
                }
            }
            return srs;
        }
        else {
            Object clientItf = comp.getFcInterface(name);
            Object sr = getServiceReference(clientItf);
            return sr;
        }
    }
    
    /**
     * Return a {@link ServiceReference} corresponding to the specified client
     * interface.
     */
    private Object getServiceReference( Object clientItf ) {
        ServiceReference<?> sr =
            ((TinfiComponentOutInterface) clientItf).getServiceReference();
        return sr;
    }
    
    /**
     * Destroy the specified content instance.
     */
    protected void destroyFcInstance( Object content ) {
                
        /*
         * Invoke the @Destroy method, if any.
         */
        if( ccmd.destroyMethod != null ) {
            try {
                ccmd.destroyMethod.invoke(content);
            }
            catch( Exception e ) {
                throw new TinfiRuntimeException(e);
            }
        }
    }

    /**
     * ThreadLocal stack holding request contexts.
     * 
     * An element is pushed on the stack by the interface (see classes generated
     * by {@link InterfaceSourceCodeGeneratorFactory}) each time an invocation
     * is received. The element is poped when the invocation has been processed
     * and just before returning the result.
     */
    protected ThreadLocal<Stack<RequestContext>> tl =
        new ThreadLocal<Stack<RequestContext>>();

    public Stack<RequestContext> getRequestContextStack() {
        return tl.get();
    }
    
    public void setRequestContextStack( Stack<RequestContext> stack ) {
        tl.set(stack);
    }    
}
